/*				11-Dec-89 10:27:59am										*/

/*--------------------------------------------------------------------------*/
/*  History: main.c, V0.0, ??-Nov-86 11:36:31am, Dan Dodge,	Baseline		*/
/*  				 V0.1, 13-Jan-87 11:36:31am, Dan Hildebrand				*/
/*----------------------------------------------------------------------------

     This file contains the command line parser and some of  the  commonly
     needed  routines  for  the various protocols that make up the Quantum
     Communications Protocol family. User written protocols should link in
     this module and include "main.h"  to  interface  cleanly  to  Quantum
     communications products.    Use  of  this command line interface also
     makes user written protocols  installable  into  the  /qterm/protocol
     file, thus  making  the  protocol  usable  from  within  Qterm. Other
     application programs ( such as bulletin board programs ) can refer to
     the /qterm/protocol file when a choice of protocol is needed.

     The  following  routines  must be defined and linked with this source
     module to produce a working protocol.    Please  refer  to  the  file
     "protocol.doc" for details.

     void help(void);				- Display help message
     int extra_option(char *);		- Process extra command line options
     void init_mdm(void);			- Initialize protocol variables
     void multi_send_check(void);	- Check if batch mode supported
     void done(int );				- Exit with status, shutdown protocol
     void final_init(void);			- Last init called before snd/rcv
     int snd(char *);				- Send a file
     receive_files( int argc, char *argv[] ) - Receive one or more files

----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include <sys/fsys.h>
#include <sys/dev.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <i86.h>
/* Give EXT a null definition so that externals become local to this module	*/
#define	EXT
#include "main.h"

unsigned old_mode;

unsigned ungot_char = 0xffff;
char device[ MAX_TTY_NAME+1 ];

void
main( int argc, char *argv[] ) {
	int file_count = 0;
	struct _dev_info_entry dinfo;
	int opt;

	if ( need_usage( argv ) || argc < 2 )
		print_usage( argv );

	prefix[0] = '\0';	/* No prefix and forced file name off by default	*/
	logfile[0] = '\0';	/* No activity log file by default */
	force = newer = relaxed = today = verbose = quiet = qterm = FALSE;
	make_dir = TRUE;
	onesec = ONESEC;
	retry_limit = RETRY_MAX;
	optind = 0;

	++optind;		/* Skip over command name	*/
	/*-- Check for a device over-ride ----*/
	if ( *argv[ optind ] == '/' ) {
		if ( (mdm = open( argv[1], O_RDWR ) ) == -1) {
			fprintf( stderr, "qcp: unable to open device '%s'.\r\012", argv[1] );
			exit( -1 );
			}
		++optind;		/* Skip over specified device name	*/
		}

	/*-- No device override so use stdin/stdout to communicate ----*/
	else {
		if ( dev_info( 0, &dinfo ) == -1 ) {
			fprintf( stderr, "qcp: Stdin not a device.\r\012" );
			exit( -1 );
			}
		strncpy( device, dinfo.tty_name, MAX_TTY_NAME );
		if ( (mdm = open( device, O_RDWR ) ) == -1) {
			fprintf( stderr, "qcp: unable to open stdin as r/w.\r\012" );
			exit( -1 );
			}
		use_stdin = TRUE;
		}

	init_mdm();			/* Setup the selected modem port and initialize		*/
						/*	any variables associated with the protocol.		*/

	if		( ( *argv[optind] | ' ' ) == 'r')					sending = FALSE;
	else if ( ( *argv[optind] | ' ' ) == 's'  &&  argc > 2 )	sending = TRUE;
	else {
		fprintf( stderr, "Invalid argument '%s', must be 'se' or 're'.\r\l",
						argv[optind] );
		exit( -1 );
		}
	++optind;

	/*---- Ready for command line options ----*/
	while( ( opt = getopt( argc, argv, "mnrtuvVQf:l:p:s:" ) ) != -1 ) {
		switch( opt ) {
			case 'f':		strcpy( prefix, optarg );
							force = TRUE;				break;
			case 'l':		strcpy( logfile, optarg );	break;
			case 'm':		make_dir = FALSE;			break;
			case 'n':		newer = TRUE;				break;
			case 'r':		relaxed = TRUE;
							onesec <<= 2;		/* Timeouts 4 times larger	*/
							retry_limit <<= 1;	/* Twice as many retries	*/
							break;
			case 's':		sub_packet_size = atoi( optarg );
							sub_packet_size >>= 3;	/* Divide by 8 for sub_packet	*/
							if ( !sub_packet_size ) {
								fprintf( stderr, "qcp: Packet size too small.\r\012" );
								exit( -1 );
								}
							if ( sub_packet_size > MAXSIZE ) {
								fprintf( stderr, "qcp: Packet size too large. Limit is %d.\r\012",
											MAXSIZE*8);
								exit( -1 );
								}
							break;
			case 't':		today = TRUE;					break;
			case 'u':		unlink_file = TRUE;				break;
			case 'V':
			case 'v':		verbose = TRUE;					break;
			case 'q':		quiet = TRUE;					break;
			case 'Q':		qterm = TRUE;					break;
			case 'p':		strcpy( prefix, optarg );		break;
			case '?': 		if ( !extra_option( (char *)&optopt ) ) {
								fprintf( stderr, "%s : Unknown option\r\012", optopt );
								exit( -1 );
								}
							break;
			default: 		++file_count;		/* Count filenames		*/
							continue;			/*	but don't null them	*/
			}
		}

	if ( file_count > 1 ) multi_send_check();	/* Check if multi-send OK	*/

	final_init();						/* Last changes before starting		*/

	if ( !use_stdin ) raw( 0 );		/* Raw console input for abort requests	*/
	if ( sending )	send_files(		argc - optind, argv + optind );
	else	 		receive_files(	argc - optind, argv + optind );

	done( 0 );				/* Protocol dependant termination routine		*/
	}

void
send_files( int argc, char *argv[] ) {
	FILE *fp;
	int i, rc;
	char	xfile_buf[110];	/* Place to read file name entries from xfile	*/

	for( i = 0; i < argc; ++i ) {
		if ( argv[i][0] == 'x'  &&  argv[i][1] == '=' ) {
			multi_send_check();		/* Does the protocol do multi-send ?	*/
			if ( !(fp = fopen( &argv[i][2], "r" ) ) ) {
				fprintf( stderr, "Unable to open index '%s'.\r\012", &argv[i][2] );
				exit( -1 );
				}
			while( fgetln( xfile_buf, 100, fp ) ) {
				rc = snd( xfile_buf );
				writelog( "SND", rc, xfile_buf );
				if( !rc )	done( -1 );
				}
			fclose( fp );
			}
		else {
			rc = snd( argv[i] );
			writelog( "SND", rc, argv[i] );
			if( !rc )	done( -1 );
			}
		}
	done( TRUE );
	}

read_mdm( int nchars, int timeout, uchar *buf ) {
	int n, i, ungot = 0;

	if ( ungot_char != 0xffff ) {		/* Process unget buffer first	*/
		--nchars;
		*buf++ = (uchar) ungot_char;
		ungot_char = 0xffff;
		ungot = 1;
		}
	if ( timeout == -1 ) timeout = 0;
/*	n = dev_read( mdm, buf, nchars, nchars, 0, timeout, 0, NULL );	*/
	n = 0;
	while ( nchars  &&
			(i = dev_read( mdm, buf, nchars, nchars, 0, timeout, 0, 0 ) ) ) {
		if ( i == -1 ) perror( "read_mdm()" );
		buf += i;
		nchars -= i;
		n += i;
		}
	return( n + ungot );
	}

void
writelog( char *str, int rc, char *msg ) {
	int i = 0;
	time_t tt;
	FILE *fp;

	if ( logfile[0] && *msg ) {
		tt = time (NULL);
		while( i++ < 5 ) {	/* Try 5 times to open the potentially busy file*/
			if ( !( fp = fopen( logfile, "a" ) ) ) {
				wait_ticks( 10 );			/* 0.5 seconds between attempts	*/
				continue;
				}
			fprintf( fp, "%s  %s %d  %s\r\012", ctime( &tt ), str, rc, msg );
			fclose( fp );
			break;
			}
		if ( i == 5 )
			fprintf( stderr, "Unable to append to Logfile '%s'.\r\012", logfile );
		}
	}

void
write_mdm( unsigned nchars, void *buf ) {
	write( mdm, buf, nchars );
	}

void
putmsg( char *fmt, ... ) {
	va_list	arglist;

	if ( quiet  ||  use_stdin ) return;
	va_start( arglist, fmt );
	vprintf( fmt, arglist );
	fflush( stdout );
	}

void
puterr( char *fmt, ... ) {
	va_list arglist;

	if ( quiet  ||  use_stdin  ||  !verbose ) return;
	va_start( arglist, fmt );
	vprintf( fmt, arglist );
	fflush( stdout );
	}

void
wait_ticks( int time ) {
	if ( time ) delay( time * 50 );		/* Delay units of 50 milliseconds 	*/
	}

abort_check() {
	char c;

	if ( use_stdin  ||  !dev_ischars( 0 ) ) return( FALSE );
	read( 0, &c, 1 );
	if ( c == ' ' || c == 27 || c == '\r' ) {
		printf( "\r\012Stop %s (y/n) ? ", sending ? "sending" : "receiving" );
		fflush( stdout );
		read( 0, &c, 1 );
		printf( "\r\012" );
		fflush( stdout );
		return( (c == 'y')  ||  (c == 'Y') );
		}
	return( FALSE );
	}

/*----------------------------------------------------------------------------
     This  routine  translates a received filename into a filename allowed
     by the current command line options. If the "-f forced_filename"  option
     is set,  the  specified  filename  will  be  placed into "new".  If a
     filename prefix was specified with the -p option, any node and  drive
     information  on  the  received  filename  will  be  removed  and  the
     specified path will be prepended to the filename in "new".   Provided
     the make_dir flag is enabled, any necessary directories to accomodate
     the new  filename  will be created. When a prefix is specified, a '^'
     in the received filename is not tolerated.

IMPORT:	old - Pointer to old filename.
EXPORT: new - Pointer to new filename.
		TRUE - Filename is valid.
		FALSE - Filename is not valid.
----------------------------------------------------------------------------*/
make_file_name( char *new, char *old ) {
	register char	*p;
	register int len = strlen( old );

	strcpy( new, prefix );					/* Set prefix into new name		*/
	if ( force ) return( TRUE );			/* If force flag, then all done	*/
	if ( !len ) return( FALSE );			/* Must be a file name			*/
	p = old;								/* Start with old name			*/
	if ( prefix[0] ) {		/* No directory climbing if a prefix was set	*/
		if ( strchr( old, '^' ) ) return( FALSE );	/* QNX 2.15 style		*/		p = old;
		while ( *p ) {
			if ( *p == '.'  &&  *(p+1) == '.' ) return( FALSE );
			++p;
			}
		p = old;					/* Start with old name					*/
		if ( *p == '[' )			/* Skip over the QNX 2.15 node number	*/
			while( *p  &&  *p++ != ']' ) ;
		if ( *p == '/'  &&  *(p+1) == '/' ) {
			p += 2;					/* Skip over the QNX4.0 node number		*/
			while( *p  &&  *p++ != '/' ) ;
			}
		if ( *p == '$' ) return( FALSE );	/* Can't prepend a device name	*/
		if ( *p && *(p+1) == ':') p += 2;	/* Skip over the drive number	*/
		if ( *p == '/' ) ++p;				/* Skip any path info			*/
		if ( *p == 0) return( FALSE );		/* Nothing left...				*/
		}
	strcat( new, p );	/* Place what's left of old name into the new name	*/
	if ( qnx_create_path_to( new ) == -1 ) return( FALSE );
	return( TRUE );
	}

/*	A useful debugging routine	*/
void
dump( char *msg, int n, uchar *buf ) {
	register int i;

	putmsg( "\r\012%s\ncount = %d\r\012 ", msg, n );
	for ( i = 0; i < n; ++i ) putmsg( " %02x", buf[i] );
	printf( "\r\012" );
	fflush( stdout );
	}

void
flush_in() {
	tcflush( mdm, TCIFLUSH );
	}

void
proto_ungetc( uchar ch ) {
	ungot_char = (unsigned) ch;
	}

char *
fgetln( char *buf, int length, FILE *fp ) {
	register int len;

	if ( fgets( buf, length, fp ) != buf ) return( NULL );
	if ( (len = strlen( buf )) > 0 ) {
		if ( buf[ len-1 ] == '\n' ) buf[ len-1 ] = '\0';
		}
	return( buf );
	}

void
raw( int fd ) {
	old_mode = dev_mode( fd, 0, _DEV_MODES );
	}

void
restore_opt( int fd ) {
	dev_mode( fd, old_mode, _DEV_MODES );
	}
